#!/usr/bin/env python
import click
import os
import steering_api
import logging
import yaml
import capschemas

log = logging.getLogger(__name__)

@click.command()
@click.argument('workdir')
@click.option('-t','--toplevel', default = os.getcwd())
@click.option('-v','--verbosity', default = 'INFO')
@click.option('-i','--loginterval', default = 30)
@click.option('-c','--schemasource', default = capschemas.schemadir)
@click.option('-b','--backend', default = 'multiproc:2')
@click.option('--interactive/--not-interactive', default = False)
@click.argument('workflow')
@click.argument('initdata', default = '')
def main(workdir,workflow,initdata,toplevel,verbosity,loginterval,schemasource,backend,interactive):
    logging.basicConfig(level = getattr(logging,verbosity))
    initdata = yaml.load(open(initdata)) if initdata else {}

    if backend.startswith('multiproc'):
        import backends.packtivitybackend as pb
        nparallel = int(backend.split(':')[1])
        backend = pb.PacktivityMultiProcBackend(nparallel)
    elif backend == 'celery':
        import backends.celeryapp
        import backends.packtivity_celery as pc
        backend = pc.PacktivityCeleryBackend(backends.celeryapp.app)

    steering_api.run_workflow(
        workdir,
        workflow,
        initdata,
        toplevel,
        loginterval,
        schemadir = schemasource, backend = backend, user_interaction = interactive)

if __name__ == '__main__':
    main()



import shlex
import pydotplus
import jsonpointer
import networkx as nx
from networkx.drawing.nx_pydot import write_dot
import logging
import jq
import subprocess

log = logging.getLogger(__name__)

def simple_stage_graph(workflow):
    graph = nx.DiGraph()
    for stage in workflow.stages.values():
        graph.add_node(stage.name)
        for x in stage.dependencies:
            graph.add_edge(x,stage.name)
    return graph

def write_stage_graph(workdir,workflow):
    graph = simple_stage_graph(workflow)
    write_dot(graph,'{}/yadage_stages.dot'.format(workdir))
    subprocess.call(shlex.split('dot -Tpdf {}/yadage_stages.dot'.format(workdir)),
                    stdout = open('{}/yadage_stages.pdf'.format(workdir),'w'))

def output_id(stepid,outputkey,index):
    identifier = 'output_{}_{}'.format(stepid,outputkey)
    if index is not None:
        identifier += '_{}'.format(index)
    return identifier

def add_outputs_to_cluster(step,cluster):
    #add outputs circles
    for k,v in step.result.iteritems():
        for i,y in (enumerate(v) if type(v)==list else [(None,v)]):
            name = output_id(step.identifier,k,i)
            label = '{}[{}]: '.format(k,i) if i is not None else '{}: '.format(k)
            label += ' {}'.format(y)
            cluster.add_node(pydotplus.graphviz.Node(name, label = label, color = 'blue'))
            cluster.add_edge(pydotplus.graphviz.Edge(step.identifier,name))
    
def add_step_to_cluster(step,workflow,stagecluster,provgraph):
    step = workflow.dag.getNode(step)
    stepid = step.identifier

    pars = step.task.attributes.copy()
    parstrings =  [': '.join((k,str(pars[k]))) for k in sorted(pars.keys())]

    step_report = u'''\
{name}
______
{pars}
'''

    rep = step_report.format(name = step.name, pars = '\n'.join(parstrings))
    stagecluster.add_node(pydotplus.graphviz.Node(
                name = stepid,
                obj_dict = None,
                color = 'red',
                label = rep,
                shape = 'box'
        )
    )
    add_outputs_to_cluster(step,stagecluster)

    #connect node to outputs
    if step.task.inputs:
        # if input information is there, add edge to input
        for reference in step.task.inputs:
            refid = reference.stepid
            refptr = reference.pointer
            index =  refptr.parts[1] if len(refptr.parts)>1 else None
            provgraph.add_edge(pydotplus.graphviz.Edge(output_id(refid,refptr.parts[0],index),stepid))
    else:
        #if not, we'll just add to the dependent node directly
        for pre in workflow.dag.predecessors(stepid):
            log.warning('really no inputs to this node but predecessors?: %s',step)
            provgraph.add_edge(pydotplus.graphviz.Edge(pre,stepid))

def fillscope(cluster,workflow,scope = ''):
    # scopecluster = stagecluster = pydotplus.graphviz.Cluster(graph_name = '_'.join(stagescopeprts),
    scopecluster = pydotplus.graphviz.Cluster(
                        graph_name = scope.replace('/',''),
                        label = ''.join(['[{}]'.format(p) for p in scope.split('/')[1:]]),
                        style = 'solid',
                        color = 'blue')
    cluster.add_subgraph(scopecluster)
    scopeptr = jsonpointer.JsonPointer(scope)
    scoped = scopeptr.resolve(workflow.stepsbystage)
    for stage,elements in scoped.iteritems():
        stagescopeprts = scopeptr.parts+[stage]
        stagecluster = pydotplus.graphviz.Cluster(
            graph_name = '_'.join(stagescopeprts),
            label = stage,
            labeljust = 'l',
            color = 'grey',
            style = 'dashed')
        scopecluster.add_subgraph(stagecluster)
        for i,element in enumerate(elements):
            if '_nodeid' in element:
                element = element['_nodeid']
                targetcl = stagecluster if stage != 'init' else scopecluster
                shape = 'diamond' if stage == 'init' else 'box'
                label = '' if stage == 'init' else '{}[{}]'.format(stage,i)
                additional = {'fixedsize':True, 'height':0.2, 'width':0.2} if stage == 'init' else {}
                targetcl.add_node(pydotplus.graphviz.Node(element, label = label, color = 'blue', shape = shape,**additional))
                add_result(targetcl,element,workflow.dag.getNode(element).result)
            elif type(element)==dict:
                fillscope(stagecluster,workflow,jsonpointer.JsonPointer.from_parts(scopeptr.parts+[stage,i]).path)


def path_to_id(stepid,path):
    return '{}_{}'.format(stepid,path.replace('/','_'))

def add_result(graph,parent,jsondata):
    allleafs = jq.jq('leaf_paths').transform(jsondata, multiple_output=True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x) for x in allleafs]

    for leaf in leafpointers:
        leafid = path_to_id(parent,leaf.path)
        # value = leaf.resolve(jsondata)
        source = ''.join(['[{}]'.format(p) for p in leaf.parts])
        label = '{}'.format(source)
        graph.add_node(pydotplus.graphviz.Node(leafid, label = label, color = 'red'))
        graph.add_edge(pydotplus.graphviz.Edge(parent,leafid))

def attach_to_results(provgraph,workflow,node):
    for dep in workflow.dag.getNode(node).task.inputs:
        resultid =  path_to_id(dep.stepid,dep.pointer.path)
        provgraph.add_edge(pydotplus.graphviz.Edge(resultid,node))
        
        
def connect(provgraph,workflow):
    for node in workflow.dag.nodes():
        attach_to_results(provgraph,workflow,node)
        
def write_prov_graph(workdir,workflow):
    provgraph = pydotplus.graphviz.Graph()

    fillscope(provgraph,workflow)
    connect(provgraph,workflow)

    with open('{}/yadage_workflow_instance.dot'.format(workdir),'w') as dotfile:
        dotfile.write(provgraph.to_string())
        
    with open('{}/yadage_workflow_instance.pdf'.format(workdir),'w') as pdffile:
        subprocess.call(shlex.split('dot -Tpdf {}/yadage_workflow_instance.dot'.format(workdir)), stdout = pdffile)



import capschemas

def workflow(source, toplevel, schema_name = 'yadage/workflow-schema', schemadir = None, validate = True):
    data = capschemas.load(source,toplevel,schema_name,schemadir,validate)
    return data


#!/usr/bin/env python

import os
import json
import yaml
import click

import adage
import adage.visualize

import yadage.backends.packtivity_celery
import yadage.backends.celeryapp
import yadage.workflow_loader
import yadage.yadagemodels
import yadage.visualize
import yadage.interactive
import logging
logging.basicConfig()

@click.group()
def mancli():
    pass

@mancli.command()
@click.argument('workdir')
@click.argument('workflow')
@click.argument('initdata', default = '')
@click.option('-t','--toplevel', default = os.getcwd())
@click.option('-s','--statefile', default = 'yadage_instance.json')
def init(workdir,workflow,initdata,statefile,toplevel):
    workflow_def = yadage.workflow_loader.workflow(
        toplevel = toplevel,
        source = workflow
    )

    rootcontext = {
        'readwrite': [os.path.abspath(workdir)],
        'readonly': []
    }

    workflow = yadage.yadagemodels.YadageWorkflow.createFromJSON(workflow_def,rootcontext)

    initdata = yaml.load(open(initdata)) if initdata else {}
    workflow.view().init(initdata)


    click.secho('initialized workflow', fg = 'green')

    yadagedir = '{}/_yadage'.format(workdir)
    os.makedirs(yadagedir)
    statefile = '{}/{}'.format(yadagedir,statefile)
    click.secho('statefile at {}'.format(statefile))
    with open(statefile,'w') as f:
        json.dump(workflow.json(),f)



def decide_rule(rule,state):
    click.secho('we could extend DAG with rule', fg = 'blue')
    click.secho('rule: {}/{} ({})'.format(rule.offset,rule.rule.name,rule.identifier))
    shall = raw_input(click.style("Shall we? (y/N) ", fg = 'blue')).lower() == 'y'
    if shall:
        click.secho('ok we will extend.', fg = 'green')
    else:
        click.secho('maybe another time...', fg = 'yellow')
    return shall

def decide_step(dag,nodeobj):
    print 'we could submit a DAG node (id: {}) DAG is: {}'.format(nodeobj,dag)
    shall = raw_input(click.style("Shall we? (y/N) ", fg = 'magenta')).lower() == 'y'
    if shall:
        click.secho('ok we will submit.', fg = 'green')
    else:
        click.secho('will not submit for now...', fg = 'yellow')
    return shall

def custom_decider(decide_func):
    # we yield until we receive some data via send()
    def decider():
        data = yield
        while True:
            data = yield decide_func(*data)
    return decider

@mancli.command()
@click.argument('workdir')
@click.option('-s','--statefile', default = 'yadage_instance.json')
def step(workdir,statefile):
    yadagedir = '{}/_yadage'.format(workdir)
    backend = yadage.backends.packtivity_celery.PacktivityCeleryBackend(
        yadage.backends.celeryapp.app
    )

    statefile = '{}/{}'.format(yadagedir,statefile)
    click.secho('loading state from {}'.format(statefile))
    workflow = yadage.yadagemodels.YadageWorkflow.fromJSON(
        json.load(open(statefile)),
        yadage.backends.packtivity_celery.PacktivityCeleryProxy,
        backend
    )

    extend_decider,submit_decider = yadage.interactive.interactive_deciders()

    coroutine = adage.adage_coroutine(backend,extend_decider,submit_decider)
    coroutine.next() #prime the coroutine....
    coroutine.send(workflow)

    try:
        click.secho('try stepping workflow')
        coroutine.next()
    except StopIteration:
        click.secho('workflow done.', fg = 'green')
        yadage.visualize.write_prov_graph(yadagedir,workflow)

    with open(statefile,'w') as f:
        json.dump(workflow.json(),f)


if __name__ == '__main__':
    mancli()



import logging
import adage.backends
import yadage.yadagestep
from packtivity import packtivity_callable
log = logging.getLogger(__name__)

class PacktivityProxyBase(object):
    '''
    A generic serializable proxy wrapper around a proxy object,
    that is passed in the ctor. Implementations can override details
    and proxyname methods
    '''
    def __init__(self,proxy):
        self.proxy = proxy

    def details(self):
        return None

    def proxyname(self):
        return 'PacktivityProxyBase'

    def json(self):
        return {
            'type':self.proxyname(),
            'proxydetails': self.details()
        }

class AdagePacktivityBackendBase(object):
    '''
    A wrapper generic backend base class around an existing adage backend.
    Implementations need to override the submit method and figure out how
    to submit the task and return a appropriate proxy object deriving from
    PacktivityProxyBase
    '''

    def __init__(self,adagebackend):
        self.adagebackend = adagebackend

    def result(self,resultproxy):
        return self.adagebackend.result(resultproxy.proxy)

    def ready(self,resultproxy):
        return self.adagebackend.ready(resultproxy.proxy)

    def successful(self,resultproxy):
        return self.adagebackend.successful(resultproxy.proxy)

    def fail_info(self,resultproxy):
        return self.adagebackend.fail_info(resultproxy.proxy)

class PacktivityMultiProcBackend(AdagePacktivityBackendBase):
    def __init__(self,nparallel = 2):
        super(PacktivityMultiProcBackend,self).__init__(adage.backends.MultiProcBackend(nparallel))

    def submit(self,task):
        '''
        if the task type is a genuine yadagestep we submit is as a packtivity callable,
        if it's an init step, which has a trivial call body, we'll just use it directly
        '''
        tasktype = type(task)
        if tasktype == yadage.yadagestep.yadagestep:
            acallable = packtivity_callable(task.name,task.spec,task.attributes,task.context)
        elif tasktype == yadage.yadagestep.initstep:
            acallable = task
        else:
            raise RuntimeError('cannot figure out how to submit a task of type {}'.format(tasktype))

        multiprocprox = self.adagebackend.submit(acallable)

        #since we can't really persistify the proxies of an in-memory process pool, we'll just return the base
        return PacktivityProxyBase(multiprocprox)



#!/usr/bin/env python
import adage
import adage.backends
import logging
import os
import json
import workflow_loader
from yadage.yadagemodels import YadageWorkflow
import visualize
import interactive


log = logging.getLogger(__name__)

def run_workflow(
    workdir,
    workflow,
    initdata,
    loadtoplevel,
    loginterval,
    schemadir,
    backend,
    user_interaction = False,
    validate = True
    ):
    """
    Main entry point to run a Yadage workflow
    """

    log.info('running yadage workflow %s',workflow)
    if not os.path.exists(workdir):
        os.makedirs(workdir)

    rootcontext = {
        'readwrite': [os.path.abspath(workdir)],
        'readonly': []
    }
    for k,v in initdata.iteritems():
        candpath = '{}/inputs/{}'.format(workdir,v)
        if os.path.exists(candpath):
            initdata[k] = '{}/inputs/{}'.format(rootcontext['readwrite'][0],v)

    workflow_json = workflow_loader.workflow(
        workflow,
        toplevel = loadtoplevel,
        schemadir = schemadir,
        validate = validate
    )
    workflow = YadageWorkflow.createFromJSON(workflow_json,rootcontext)
    workflow.view().init(initdata)

    yadagedir = '{}/_yadage/'.format(workdir)
    os.makedirs(yadagedir)

    with open('{}/yadage_instance_before.json'.format(yadagedir),'w') as f:
        json.dump(workflow.json(),f)

    if user_interaction:
        extend,submit = interactive.interactive_deciders()
        interactive_kwargs = {
            'extend_decider': extend,
            'submit_decider': submit
        }
    else:
        interactive_kwargs = {}

    adage.rundag(workflow,
                 track = True,
                 backend = backend,
                 update_interval = 0.02,
                 trackevery = loginterval,
                 workdir = '{}/_adage'.format(workdir),
                 **interactive_kwargs
                )

    with open('{}/yadage_instance.json'.format(yadagedir),'w') as f:
        json.dump(workflow.json(),f)
    with open('{}/yadage_template.json'.format(yadagedir),'w') as f:
        json.dump(workflow_json,f)

    # import IPython
    # IPython.embed()

    visualize.write_prov_graph(yadagedir,workflow)
    log.info('finished yadage workflow %s',workflow)



import logging
import utils
import jsonpointer
import itertools
import os
import copy
import jq
import jsonpath_rw

from yadage.yadagestep import yadagestep, initstep, outputReference
from yadage.yadagemodels import jsonStage

log = logging.getLogger(__name__)

handlers,scheduler = utils.handler_decorator()

### A scheduler does the following things:
###   - attached new nodes to the DAG
###   - for each added step
###     - the step is given a name
###     - the step attributes are determined using the scheduler spec and context
###     - a list of used inputs (in the form of [stepname,outputkey,index])

def pointerize(jsondata, asref = False, stepid = None):
    allleafs = jq.jq('leaf_paths').transform(jsondata, multiple_output=True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x).path for x in allleafs]
    jsondata_proxy = copy.deepcopy(jsondata)
    for leaf in leafpointers:
        x = jsonpointer.JsonPointer(leaf)
        x.set(jsondata_proxy, outputReference(stepid,x) if asref else x.path)
    return jsondata_proxy

def select_reference(step,selection):
    pointerized = pointerize(step.result, asref = True, stepid = step.identifier)
    matches = jsonpath_rw.parse(selection).find(pointerized)
    if not matches:
        log.error('no matches found for selection %s in result %s',selection,step.result)
        raise RuntimeError('no matches found in reference selection')


    if len(matches) > 1:
        log.error('found multiple matches to query: %s within result: %s\n \ matches %s',selection,step.result,matches)
        raise RuntimeError('multiple matches in result jsonpath query')
    return matches[0].value

def combine_outputs(outputs,flatten,unwrapsingle):
    combined = []
    for reference in outputs:
        if type(reference)==list:
            if flatten:
                for elementref in reference:
                    combined+=[elementref]
            else:
                combined+=[reference]
        else:
            combined+=[reference]
    if len(combined)==1 and unwrapsingle:
        combined = combined[0]
    return combined

def select_steps(stage,query):
    return stage.view.getSteps(query)

def select_outputs(steps,selection,flatten,unwrapsingle):
    return combine_outputs(map(lambda s: select_reference(s,selection),steps),flatten,unwrapsingle)

def resolve_reference(stage,selection):
    if type(selection) is not dict:
        return None
    else:
        steps   = select_steps(stage, selection['stages'])
        outputs = select_outputs(steps, selection['output'], selection.get('flatten',False), selection.get('unwrap',False))
        return outputs

def select_parameter(stage,parameter):
    if type(parameter) is not dict:
        value = parameter
    else:
        value = resolve_reference(stage,parameter)
    return value

def finalize_value(stage,step,value,context):
    if type(value)==outputReference:
        step.used_input(value)
        v = value.pointer.resolve(stage.view.dag.getNode(value.stepid).result)
        return finalize_value(stage,step,v,context)
    if type(value)==list:
        return [finalize_value(stage,step,x,context) for x in value]
    if type(value) in [str,unicode]:
        return value.format(**context)
    return value

def finalize_input(stage,step,json,context):
    context = context.copy()
    context['workdir'] = context['readwrite'][0]
    result = {}
    for k,v in json.iteritems():
        if type(v) is not list:
            result[k] = finalize_value(stage,step,v,context)
        else:
            result[k] = [finalize_value(stage,step,element,context) for element in v]
    return result

def step_or_init(name,spec,context):
    if 'step' in spec:
        return yadagestep(name = name, spec = spec['step'], context = context)
    elif 'workflow' in spec:
        return initstep('init {}'.format(name))

def make_new_context(name,oldcontext):
    newcontext = {'readwrite':['{}/{}'.format(oldcontext['readwrite'][0],name)], 'readonly':[]}
    newcontext['readonly'] += [ro for ro in itertools.chain(oldcontext['readonly'],oldcontext['readwrite'])]
    os.makedirs(newcontext['readwrite'][0])
    return newcontext

def addStepOrWorkflow(name,stage,step,spec):
    if type(step)==initstep:
        newcontext = make_new_context(name,stage.context)
        subrules = [jsonStage(yml,newcontext) for yml in spec['workflow']['stages']]
        stage.addWorkflow(subrules, initstep = step)
    else:
        stage.addStep(step)

@scheduler('singlestep-stage')
def simple_stage(stage,spec):

    parameters = {
        k:select_parameter(stage,v) for k,v in spec['parameters'].iteritems()
    }

    step = step_or_init(name = stage.name, spec = spec, context = stage.context)
    finalized = finalize_input(stage,step,parameters,stage.context)

    addStepOrWorkflow(stage.name,stage,step.s(**finalized),spec)

def scatter(parameters,scatter):
    commonpars = parameters.copy()
    to_scatter = {}
    for scatpar in scatter['parameters']:
        to_scatter[scatpar] = commonpars.pop(scatpar)

    singlesteppars=[]
    if scatter['method']=='zip':
        keys, zippable = zip(*[(k,v) for i,(k,v) in enumerate(to_scatter.iteritems())])

        for zipped in zip(*zippable):
            individualpars = dict(zip(keys,zipped))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]

    if scatter['method']=='cartesian':
        for what in itertools.product(*[to_scatter[k] for k in scatter['parameters']]):
            individualpars = dict(zip(scatter['keys'],what))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]
    return singlesteppars

@scheduler('multistep-stage')
def multi_stage(stage,spec):
    parameters = {
        k:select_parameter(stage,v) for k,v in spec['parameters'].iteritems()
    }
    singlesteppars = scatter(parameters,spec['scatter'])

    for i,pars in enumerate(singlesteppars):
        index_context = stage.context.copy()
        index_context.update(index = i)

        singlename = '{}_{}'.format(stage.name,i)
        step = step_or_init(name = singlename, spec = spec, context = stage.context)
        finalized = finalize_input(stage,step,pars,index_context)

        addStepOrWorkflow(singlename,stage,step.s(**finalized),spec)



import click

def decide_rule(rule,state):
    click.secho('we could extend DAG with rule', fg = 'blue')
    click.secho('rule: {}/{} ({})'.format(rule.offset,rule.rule.name,rule.identifier))
    shall = raw_input(click.style("Shall we? (y/N) ", fg = 'blue')).lower() == 'y'
    if shall:
        click.secho('ok we will extend.', fg = 'green')
    else:
        click.secho('maybe another time...', fg = 'yellow')
    return shall

def decide_step(dag,nodeobj):
    click.echo('we could submit a DAG node (id: {}) DAG is: {}'.format(nodeobj,dag))
    shall = raw_input(click.style("Shall we? (y/N) ", fg = 'magenta')).lower() == 'y'
    if shall:
        click.secho('ok we will submit.', fg = 'green')
    else:
        click.secho('will not submit for now...', fg = 'yellow')
    return shall

def custom_decider(decide_func):
    # we yield until we receive some data via send()
    def decider():
        data = yield
        while True:
            data = yield decide_func(*data)
    return decider

def interactive_deciders():
    '''
    returns a tuple (extend,submit) of already-primed deciders for both
    extension and submission
    '''
    extend_decider = custom_decider(decide_rule)()
    extend_decider.next() #prime decider

    submit_decider = custom_decider(decide_step)()
    submit_decider.next() #prime decider

    return extend_decider,submit_decider



from celery import Celery
import os
redis_string = 'redis://{hostname}:{port}/{database}'.format(
  hostname = os.environ.get('YADAGE_REDIS_HOST','localhost'),
  port = os.environ.get('YADAGE_REDIS_PORT',6379),
  database = os.environ.get('YADAGE_REDIS_DB',0)
)
app = Celery(broker = redis_string, backend = redis_string)
